
#include "WidgetDriver.h"
#include "WidgetMessageFilter.h"
#include "../Thunk/ProcThunk.h"
#include <Windows.h>
#ifdef _DEBUG
#include <set>
#endif // _DEBUG

namespace ghost{

	namespace widget{

		class DriverImpl{
			window::ProcThunk thunk_;
			HWND hContainerWnd_;
			WNDPROC originalProc_;

		public:
			explicit DriverImpl(HWND hWnd)
				: hContainerWnd_(hWnd)
				, originalProc_(0)
			{
#ifdef _DEBUG
				assert(::IsWindow(hContainerWnd_));
				// ܶͬһ
				assert(GetContainerWindows_().insert(hContainerWnd_).second);
#endif // _DEBUG
				assert(thunk_.Init(StartWndProc_, this));
				originalProc_ = (WNDPROC)::SetWindowLong(hContainerWnd_, GWL_WNDPROC, (LONG)thunk_.GetPROC());
			}
			~DriverImpl()
			{
#ifdef _DEBUG
				GetContainerWindows_().erase(hContainerWnd_);
#endif // _DEBUG
				::SetWindowLong(hContainerWnd_, GWL_WNDPROC, (LONG)originalProc_);
			}

		private:
			DriverImpl(const DriverImpl&);
			DriverImpl& operator =(const DriverImpl&);

		public:
			HWND GetContainerWindow() const{return hContainerWnd_;}

		private:
			LRESULT WndProc_(UINT message, WPARAM wParam, LPARAM lParam)
			{
				// Ϣ
				MessageFilter::Param param;
				param.hWnd = hContainerWnd_;
				param.originalProc = originalProc_;
				param.message = message;
				param.wParam = wParam;
				param.lParam = lParam;
				return MessageFilter::Filter(param);
			}

		private:
			// ڳʼthunkĴڹ
			static LRESULT CALLBACK StartWndProc_(
				HWND hWnd, 
				UINT message, 
				WPARAM wParam, 
				LPARAM lParam)
			{
				DriverImpl* pThis = (DriverImpl*)hWnd;
				return pThis->WndProc_(message, wParam, lParam);
			}

#ifdef _DEBUG
			static std::set<HWND>& GetContainerWindows_()
			{
				static std::set<HWND> wnds;
				return wnds;
			}
#endif // _DEBUG
		};

		//--------------------------------------

		Driver::Driver(HWND hWnd)
			: pImpl_(new DriverImpl(hWnd))
		{
			
		}
		Driver::~Driver()
		{
			delete pImpl_;
		}

		inline HWND Driver::GetContainerWindow() const
		{
			return pImpl_->GetContainerWindow();
		}

	} // namespace widget

} // namespace ghost